/* 
    Artificial data generator, according to the domain specifications 
    it generates a data set.
    Usage : <domain> [-mixup] [-noise <n>] [-u <n>] [-irrelevant <n>]
    <domain> : name of the specification file
    options :
       -mixup : mixup the example orders (if seed is given it is used)
       -noise : command line noise parameter <n>:[0.0,1.0]
       -u     : command line Unknown attribute ratio <n>:[0.0,1.0]
       -irrelevant : # irrelevant attributes <n>:integer >=0
    Outputs : 
       <domain>.data : contains generated examples
       <domain>.cfp  : setting file for CFP for <domain>.data data set
*/

#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <sys/time.h>
#include <stdlib.h>

#define IRRELEVANT_MIN    0.0
#define IRRELEVANT_MAX   10.0
#define MAX_FEATURES     40   /* maximum # of features */
#define MAX_CLASSES      25   /* maximum # of classes */
#define MAX_EXAMPLES   1000   /* maximum # of examples in training set */
#define MAX_RECTANGLES  100   /* maximum # of hyperrectangles for each class */
#define STR_SIZE        300   /* maximum string size */
#define FALSE             0
#define TRUE              1
#define DISCRETE          0   /* type of the feature value */
#define CONTINUOUS        1   /* type of the feature value */
#define GEN_RATIO        10.0

extern long lrand48();
extern double drand48();
float lrand_uniform();
float drand_uniform();

typedef struct {
    int type;
    float min,max;
    float (*proc)();
}Feature;

typedef struct {
    float lower,upper;
}Range;

typedef struct {
     int n;              /* # example for the hyperrectangle */
     Range f[MAX_FEATURES];
}Rectangle;

typedef struct {
     int n,              /* # example for the class */
         nr;             /* # hyperrectangles in the class */
     Rectangle r[MAX_RECTANGLES];
}Class;

typedef struct {
     float f[MAX_FEATURES];
     int class;
}Example;

Feature features[MAX_FEATURES];
Class   classes[MAX_CLASSES+1];
Example examples[MAX_EXAMPLES];
int     exampleindexlist[MAX_EXAMPLES];
int  correct_classes[MAX_EXAMPLES];
int nRun=1, CrossValidation=FALSE,Randomize=FALSE;

int nFeatures,nIrrelevants,nExamples,nClasses,nConcepts,Seed;
float NoiseLevel;  /* rate of the error to be introduced to the class values */
float UnknownAttributeRatio; /* ratio of the unknown attribute values */
int Mixup;           /* rearrange order of the examples */
char domain[STR_SIZE], examplefile[STR_SIZE],cfpfile[STR_SIZE];

error_exit(n,msg)
int n;
char *msg;
{
  printf("Error %d: %s\n",n,msg);
  exit(n);
}

init_seed()
{
  unsigned short dseed[3];
  struct timeval now;
  struct timezone tz;
  if (Seed <=0 ) {
    gettimeofday(&now,&tz);
    dseed[0]=(unsigned short) now.tv_usec;
    dseed[1]=(unsigned short) now.tv_sec;
    dseed[2]=(unsigned short) now.tv_usec;
  } 
  else {dseed[0]=dseed[1]=dseed[2]=(unsigned short) Seed; }
  seed48(dseed);
}

/* reorder orgexample into newexample (with n examples and na attributes)*/
mixup(list,n)
int *list,n;
{
  int i, x;
  static int dist[MAX_EXAMPLES];

  for (i=0; i<n; i++)
    dist[i]=0;
  i=0;
  while (1) {
    if (i>=n)
      break;
    x=lrand48()%n;
    if (dist[x] < 1) {
      list[i]=x;
      dist[x]++;
      i++;
    }
  }
}

float drand_uniform(lower,upper)
float lower,upper;
{
  return((float)lower+(upper-lower)*drand48());
}

float lrand_uniform(lower,upper)
float lower,upper;
{  
  return((int)drand_uniform(lower,upper));
}

float rand_uniform(lower,upper,F)
float lower,upper;
int F;
{
    return((*features[F].proc)(lower,upper));
}

/* read domain specification */
read_specification(fn)
char *fn;
{
  FILE *fp;
  int i,j,c,n;
  char temp[20],line[STR_SIZE],str[STR_SIZE], *cp,*cp2,*cpos;

  fp=fopen(fn,"r");
  if (fp==NULL)
    error_exit(1,"specification file is not found...");
  
  for (cp=fgets(line,STR_SIZE,fp); cp !=NULL; cp=fgets(line,STR_SIZE,fp)) {
    if (sscanf(line,"%s",str) <1) continue; /* ignore empty line */
    if (line[0]==';') continue;	/* ignore comment line */
    if (stricmp(str,"Noise")==0)  {
      if (sscanf(line,"%s%f", str, &NoiseLevel) <2)
	error_exit(2,"Settings file must specify Noise level");
    }
    else if (stricmp(str,"UnknownAttributeRatio")==0)  {
      if (sscanf(line,"%s%f", str, &UnknownAttributeRatio) <2)
	error_exit(2,"Settings file must specify UnknownAttributeRatio");
    }
    else if (stricmp(str,"Randomize")==0)  {
      if (sscanf(line,"%s%d", str, &nRun) <2)
	error_exit(2,"Settings file must specify # of Randomized runs");
      Randomize=TRUE;
    }
    else if (stricmp(str,"CrossValidation")==0)  {
      if (sscanf(line,"%s%d", str, &nRun) <2)
	error_exit(2,"Settings file must specify # of runs");
      CrossValidation=TRUE;
    }
    else if (stricmp(str,"Seed")==0)  {
      if (sscanf(line,"%s%d", str, &Seed) <2)
	error_exit(2,"Settings file must specify Randomize Seed");
    }
    else if (stricmp(str,"#features")==0)  {
      if (sscanf(line,"%s%d", str, &nFeatures) <2)
	error_exit(2,"Settings file must specify # of features");
      if (nFeatures > MAX_FEATURES)
	error_exit(3,"Too many features... needs recompiling...");
      for (i=0; i< nFeatures; i++)  {
	if (fscanf(fp,"%s %s %f %f",str,temp,
		   &features[i].min,&features[i].max)<4)
	  error_exit(3,"Setting file must specify, feature type min max ");
	if (features[i].max==features[i].min)
	  features[i].max++;
	if (stricmp(temp,"discrete")==0)
	  features[i].type=DISCRETE;
	else features[i].type=CONTINUOUS;
      }
    }
    else if (stricmp(str,"#classes")==0)  {
      if (sscanf(line,"%s%d", str, &nClasses) <2)
	error_exit(2,"Settings file must specify #classes");
      if (nClasses > MAX_CLASSES)
	error_exit(3,"Too many classes... needs recompiling...");
    }
    else if (stricmp(str,"rectangle")==0)  {
      if (sscanf(line,"%s%d%d", str, &c, &n) <3)
	error_exit(2,
		   "Settings file must specify class value and #examples for the rect.");
      if (n < 0 )
	error_exit(3,"Too few examples required for this rectangle");
      if (c<1 || c> nClasses)
	error_exit(3,
		   "Setting file must specify, corresponding class[1..#classes]");
      nExamples +=n;
      if (nExamples >= MAX_EXAMPLES)
	error_exit(3,"Too many examples... needs recompiling...");
      if (classes[c].nr >= MAX_RECTANGLES)
	error_exit(3,
		   "Too many hyperrectagles for this class... needs recompiling...");
      classes[c].r[ classes[c].nr ].n =n;
      n=classes[c].nr;
      for (i=0; i< nFeatures; i++)  
	if (fscanf(fp,"%f %f",&classes[c].r[n].f[i].lower,
		   &classes[c].r[n].f[i].upper)<2)
	  error_exit(3,"Setting file must specify, lower and upper bonderies for each feature");
      classes[c].nr++;
    }
  }
  fclose(fp);
}

init_globals()
{
  nExamples=0;
  nFeatures=0;
  nIrrelevants=0;
  nConcepts=0;
  NoiseLevel=0;
  UnknownAttributeRatio=0;
  Mixup=FALSE;
  memset(classes,0,sizeof(classes));
  Seed=0;
}

generate_examples()
{
  int i,n,f,k,j;

  for (j=0,i=1; i<=nClasses; i++) {
    for (k=0; k<classes[i].nr; k++) {
      for (n=0; n<classes[i].r[k].n; n++) {
	for (f=0; f<nFeatures; f++)
	  examples[j].f[f]=rand_uniform(classes[i].r[k].f[f].lower,
					classes[i].r[k].f[f].upper,f);
	examples[j].class=i;
	j++;
      }
    }
  }    
  n=nFeatures+nIrrelevants;
  for (f=nFeatures; f<n; f++) 
    for (j=0; j<nExamples; j++)
      examples[j].f[f]=drand_uniform(IRRELEVANT_MIN,IRRELEVANT_MAX);
}

add_noise()
{
  int i,n,c,E,actual;

  for (i=0; i<nExamples; i++)
    correct_classes[i]=examples[i].class;

  n=NoiseLevel*nExamples;
  for (actual=0; actual<n; ) {
    E=lrand48()%nExamples;
    c=1+lrand48()%nClasses;
    if (c!=correct_classes[E]) 
      actual++;
    examples[E].class=c;
  }
  printf("Desired Noise Level= %f, Actual= %f\n",NoiseLevel,
	 (float)actual/nExamples);
}

write_examples(fn,list)
char *fn;
int *list;
{
  FILE *fp;
  int i,j,n,k,E;

  n=0;				/* # unknown attributes values */
  fp=fopen(fn,"w");
  if (fp==NULL) error_exit(2,"Data file creation failed");
  printf("Will write the data to the file %s ...\n", fn);
  for (i=0; i< nExamples; i++) {
    E=list[i];
    for (k=0,j=0; j<nFeatures + nIrrelevants; j++)  {
      if (k<(nFeatures+nIrrelevants-1) && drand48()<UnknownAttributeRatio) {
	fprintf(fp,"%c,",'?');
	n++; 
	k++;
      }
      else if (j<nFeatures) {
	if (features[j].type==DISCRETE)
	  fprintf(fp,"%d,",(int)examples[E].f[j]);
	else
	  fprintf(fp,"%f,",examples[E].f[j]);
      }
      else fprintf(fp,"%f,",examples[E].f[j]); /* irrelevant attr. */
    }
    fprintf(fp,"%d\n",examples[E].class);
  }
  fclose(fp);
  if (UnknownAttributeRatio>0) 
    printf("Desired Unknown Attribute Level= %f, Actual= %f\n",
	   UnknownAttributeRatio,(float)n/((nFeatures+nIrrelevants)*nExamples));

}

write_cfpfile(fn)
char *fn;
{
  FILE *fp;
  int n,i;
    
  fp=fopen(fn,"w");
  if (fp==NULL) error_exit(2,"CFP setting file creation failed");
  printf("Will write the CFP settings to the file %s ...\n", fn);
  fprintf(fp,"; CFP setting file for %s dataset\n",domain);
  fprintf(fp,"#features %d\n",nFeatures+nIrrelevants);
  fprintf(fp,"#classes %d\n",nClasses);
  fprintf(fp,"classes");
  for (i=1; i <= nFeatures+nIrrelevants; i++)
    fprintf(fp, " %d", i);
  fprintf(fp, "\n");
  fprintf(fp,"; Generalization Limits for each feature\n");
  fprintf(fp,"Limits ");
  for (i=0; i<nFeatures; i++) 
    fprintf(fp,"%6.3f ",(features[i].max-features[i].min)/GEN_RATIO);
  for (i=0; i<nIrrelevants; i++) 
    fprintf(fp,"%6.3f ",(IRRELEVANT_MAX-IRRELEVANT_MIN)/GEN_RATIO);
  fprintf(fp,"\n");
  fprintf(fp,"; Feature Weight Adjustment rate\n");
  fprintf(fp,"delta_f 0.05\n");
  fclose(fp);
} 

print_rectangles()
{
  int i,j,k;

  for (j=1; j<=nClasses; j++) {
    printf("Given Hyperrectangles for Class %d\n",j);
    printf("class %d has %d rectangles\n",j,classes[j].nr);
    for (k=0; k<classes[j].nr; k++) {
      printf("class %d rectangle %d has %d examples\n",
	     j,k+1,classes[j].r[k].n);
      for (i=0; i<nFeatures; i++)  {
	if (features[i].type==DISCRETE)
	  printf("[%6d,%6d] ",(int)classes[j].r[k].f[i].lower,
		 (int)classes[j].r[k].f[i].upper);
	else
	  printf("[%6.2f,%6.2f] ",classes[j].r[k].f[i].lower,
		 classes[j].r[k].f[i].upper);
      }
      printf("%3d\n",j);
    }
  }
  printf("Total number of examples %d\n",nExamples);
}

parse_options(argc,argv)
int argc;
char *argv[];
{
  int i,n;
  float x;

  for (i=2; i<argc; i++) {
    if (stricmp(argv[i],"-m")==0) 
      Mixup=TRUE;
    else if (stricmp(argv[i],"-n")==0) {
      if (argv[i+1] !=NULL) {
	x=atof(argv[i+1]);
	if (x>=0 && x<=1)
	  NoiseLevel=x;
      }
      else error_exit(2,"Specify noise level");
    }
    else if (stricmp(argv[i],"-u")==0) {
      if (argv[i+1] !=NULL) {
	x=atof(argv[i+1]);
	if (x>=0 && x<=1)
	  UnknownAttributeRatio=x;
      }
      else error_exit(2,"Specify Unknown attribute ratio ");
    }
    else if (stricmp(argv[i],"-i")==0) {
      if (argv[i+1] !=NULL) {
	n=atoi(argv[i+1]);
	if (n>=0 )
	  nIrrelevants=n;
      }
      else error_exit(2,"Specify # irrelevant attributes ");
    }
  }
}

main(argc,argv)
int argc;
char *argv[];
{
  int i;

  if (argc < 2) {
    printf("Usage : %s <domain> [-m] [-n <x>] [-u <x>] [-i <n>]\n",argv[0]);
    printf("-m: mixup examples\n");
    printf("-n: add noise 0-1\n");
    printf("-u: add unknown attributes 0-1\n");
    printf("-i: add irrelevant attributes 0..n\n");
    return(-1);
  }
  init_globals();
  strcpy(domain,argv[1]);
  read_specification(domain);
  if (argc>2) 
    parse_options(argc,argv);
  if ((nFeatures+nIrrelevants) > MAX_FEATURES)
    error_exit(3,"Too many features... needs recompiling...");
  strcpy(examplefile,domain);
  strcat(examplefile,".data");
  strcpy(cfpfile,domain);
  strcat(cfpfile,".cfp"); 
  init_seed();
  for (i=0; i<nFeatures; i++) {
    if (features[i].type==DISCRETE)
      features[i].proc=lrand_uniform;
    else
      features[i].proc=drand_uniform;
  }
  print_rectangles();
  generate_examples();
  if (NoiseLevel > 0)
    add_noise();
  if (Mixup==TRUE)
    mixup(exampleindexlist,nExamples);
  else for (i=0; i<nExamples; i++) exampleindexlist[i]=i;
  write_examples(examplefile,exampleindexlist);
  write_cfpfile(cfpfile);
}
