/* ga-cfp.c: GA-CFP generates a <domain>.cfp file */

#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include "ga-cfp.h"

extern float pc, pm; /* ga.c */
extern int No_Chromosomes, No_Generations; /* ga.c */

int No_Features, No_Classes;
int No_Trainers, No_Testers, Train_correct, Test_correct;
float weight[MAX_FEATURES], gen_limit[MAX_FEATURES];
int linear[MAX_FEATURES];
float delta;
int className[MAX_CLASSES], classIndex[MAX_CLASSES];
float vote[MAX_CLASSES];
int LEARN_WEIGHTS;

INSTANCE trainData[MAX_DATA], testData[MAX_DATA];

char Domain[MAX_NAME], domainFile[MAX_NAME],
     trainDataFile[MAX_NAME], testDataFile[MAX_NAME],
     logFile[MAX_NAME];

int verbosity = 0;
FILE *fopen(), *df, *logf;

struct timeval *ga_start, *ga_end;
struct timezone *tz;
int ga_time;

parse_options(argc,argv)
int argc;
char *argv[];
{
  int i;

  strcpy(Domain, argv[1]);
  strcpy(domainFile, Domain);
  strcat(domainFile, ".cfp");
  strcpy(trainDataFile, Domain);
  strcat(trainDataFile, ".Train");
  i = 2;
  while (i < argc) {
    if (stricmp(argv[i], "-pc")==0)
      sscanf(argv[++i], "%g", &pc);
    else if (stricmp(argv[i], "-pm")==0)
      sscanf(argv[++i], "%g", &pm);
    else if (stricmp(argv[i], "-size")==0) 
      sscanf(argv[++i], "%d", &No_Chromosomes);
    else if (stricmp(argv[i], "-gen")==0)
      sscanf(argv[++i], "%d", &No_Generations);
    else if (strcmp(argv[i], "-v")==0) {
      strcpy(logFile, Domain);
      strcat(logFile, ".log");
      verbosity = 1;
    }
    else {
      printf("Unknown command line argument %s\n", argv[i]);
      exit (0);
    }
    i++;
  }
}

void initialize()
{
  No_Classes = No_Features = No_Testers = No_Testers = 0;
  read_domain(domainFile);
  read_data(trainDataFile, trainData, &No_Trainers);
  if (verbosity > 0) {
    logf = fopen(logFile, "a");
    fprintf(logf, "Started logging with verbosity %d.\n", verbosity);
  }
}

CLASS prediction_on_feature (f, value)
FEATURE f;
VALUE value;
{
  SEGMENT *seg, *s;

  seg = search (f, value);

  if (value < seg->upper)       /* in a range partition */
    return(seg->class);
  
  /* return the class of the segment with maximum repr */
  for (s = seg->right; (s->right !=NULL && s->lower <= value); s = s->right) 
    if (s->repr > seg->repr)
      seg = s;

  return (seg->class);
}

int classification (inst)
INSTANCE inst;
{
  CLASS c, pof, winner;
  FEATURE f;

  for (c=0; c <= No_Classes; c++) vote[c] = 0.0;

  for (f=0; f < No_Features; f++) 
    if (inst.status[f] == KNOWN) {
      pof = prediction_on_feature (f, inst.value[f]);
      vote[pof] = vote[pof] + weight[f];
    }

  /* return the class that received maximum votes */
  winner = 1;
  for (c=2; c <= No_Classes; c++) 
    if (vote[c] > vote[winner]) 
      winner = c;
  if (vote[winner] == 0.0) 
    winner = UNDETERMINED;

  return(winner);
}

float fitness(chromosome)
float *chromosome;
{
  int i, f, pof; 
  int half, correct1=0, correct2=0;

  half = No_Trainers / 2;
  delta = chromosome[0];
  for (f=0; f < No_Features; f++) gen_limit[f] = chromosome[f+1];

  /* train on the first half, test on second half */
  initialize_partitions ();
  for (f=0; f < No_Features; f++) weight[f] = 1.0;  
  for (i=0; i < half; i++)	/* train in the first half */
    for (f=0; f < No_Features; f++) 
      if (trainData[i].status[f] == KNOWN) {
	pof = prediction_on_feature (f, trainData[i].value[f]);
	if (pof == trainData[i].class) weight[f] *= (1+delta);
	else weight[f] *= (1-delta);
	update_feature_partitioning (f, trainData[i]);
      }
  for (i=half; i < No_Trainers; i++) /* test in the second half */
    if (classification(trainData[i]) == trainData[i].class) correct1++;
  free_partitions ();

  /* train on second half, test on first half */
  initialize_partitions ();
  for (f=0; f < No_Features; f++) weight[f] = 1.0;  
  for (i=half; i < No_Trainers; i++) /* train in the second half */
    for (f=0; f < No_Features; f++) 
      if (trainData[i].status[f] == KNOWN) {
	pof = prediction_on_feature (f, trainData[i].value[f]);
	if (pof == trainData[i].class) weight[f] *= (1+delta);
	else weight[f] *= (1-delta);
	update_feature_partitioning (f, trainData[i]);
      }
  for (i=0; i < half; i++)	/* test in the first half */
    if (classification(trainData[i]) == trainData[i].class) correct2++;
  free_partitions ();

  return((float)(correct1+correct2)/No_Trainers);
}

main(argc, argv)
int argc;
char *argv[];
{
  float accuracy;
  int f;

  ga_start = (struct timeval * ) malloc(sizeof(struct timeval));
  ga_end = (struct timeval * ) malloc(sizeof(struct timeval));
  tz    = (struct timezone * ) malloc(sizeof(struct timezone)); 

  if (argc == 1) {
    printf("ga-cfp domain [-v] [-pc <pc>] [-pm <pm>] [-size <population size>] [-gen <generations>] -v\n");
    exit(1);
  }
  else {
    set_ga_defaults ();
    parse_options(argc,argv);
  }

  initialize();

  if (verbosity > 0) {
    fprintf(logf, "Pop. size: %d, Generations: %d, pc: %g, pm: %g\n",
	    No_Chromosomes, No_Generations, pc, pm);
    fprintf(logf, "No_Trainers: %d\n", No_Trainers);
  }

  gettimeofday(ga_start, tz);

  accuracy = ga_learn_parameters (&delta, gen_limit);

  gettimeofday(ga_end, tz);

  /* timing */
  ga_time = (ga_end->tv_sec * 1000 - ga_start->tv_sec * 1000) +
    (ga_end->tv_usec / 1000 - ga_start->tv_usec / 1000);
  
  df = fopen(domainFile, "a");
  fprintf(df, "Delta %g\n", delta);
  fprintf(df, "Limits ");
  for (f=0; f < No_Features; f++)
    fprintf(df, "%g ", gen_limit[f]);
  fprintf(df, "\n");
  fprintf (df, "; These settings are obtained by ga-cfp with the following ga parameters:\n");
  fprintf (df, "; pc: %g, pm: %g, populations size: %d, number of generations: %d\n", pc, pm, No_Chromosomes, No_Generations);
  fprintf (df, "; Training file %s contained %d examples.\n",
	   trainDataFile, No_Trainers);
  fprintf (df, "; Train accuracy obtained is %g\n", accuracy);
  fprintf (df, "; Determining these parameters took %d msec's.\n",ga_time);
  fclose(df);
  return 0;
}
